import random
import numpy as np
import copy

__all__ = (
    'OUprocess',
)

class OUprocess:
    """Ornstein-Uhlenbeck process."""
    def __init__(self, pi, size, seed, mu=0., theta=0.15, sigma=0.2):
        self.pi = pi
        self.mu = mu * np.ones(size)
        self.theta = theta
        self.sigma = sigma
        self.seed = random.seed(seed)
        self.reset()

    def reset(self):
        self.state = copy.copy(self.mu)

    def sample(self):
        x = self.state
        dx = self.theta * (self.mu - x) + self.sigma * np.array([random.random() for i in range(len(x))])
        self.state = x + dx
        return self.state

    def act(self, s, explore=True):
        #print("Action : ", end=" ")
        noise = self.sample() if explore else 0.0
        action = self.pi.act(s) + noise
        return action


import numpy as np
from collections import deque
from itertools import islice
from .transition import TransitionBatch

__all__ = (
    'BaseRecorder',
    'NStepRecorder'
)

class BaseRecorder:
    def reset(self):
        raise NotImplementedError
    def add(self, s, a, r, done, s_next, w=1.0):
        raise NotImplementedError
    def pop(self):
        raise NotImplementedError
    def flush(self):
        if not self:
            raise ValueError("Empty recorder")
        while self:
            self.pop()
    def __bool__(self):
        raise NotImplementedError
    
class NStepRecorder(BaseRecorder):
    def __init__(self, n=1, gamma=0.9, bootstrap=False):
        super().__init__()
        self.n = n
        self.gamma = gamma
        self.bootstrap = bootstrap
        self.reset()

    def reset(self):
        self.deque_s = deque([])
        self.deque_r = deque([])
        self.done = False
        
    def __len__(self):
        return len(self.deque_s)

    def add(self, s, a, r, done, s_next):
        if self.done and len(self):
            raise ValueError("Episode completed : flush before adding")
        self.deque_s.append( (s,a,r,done,s_next) )
        self.deque_r.append( r )
        self.done = bool(done)

    def pop(self):
        if not self:
            raise ValueError("Empty recorder")

        s,a,r,done,s_next = self.deque_s.popleft()

        if self.bootstrap:
            raise NotImplementedError("bootstrapping rewards not yet implemented")
        else:
            _ = self.deque_r.popleft()

        return TransitionBatch.from_single(
            s=s, a=a, r=r, done=done, s_next=s_next, gamma=self.gamma
        )

    def popN(self, n):
        if not self:
            raise ValueError("Empty recorder")
        if n > len(self):
            raise ValueError("Not enough transitions")
        
        transitions = [self.pop() for _ in range(n)]
        cat = lambda *x: np.concatenate(x, axis=0)
        return TransitionBatch(
            S=cat([t.S for t in transitions]),
            A=cat([t.A for t in transitions]),
            R=cat([t.R for t in transitions]),
            Done=cat([t.Done for t in transitions]),
            S_next=cat([t.S_next for t in transitions]),
            Gamma=cat([t.Gamma for t in transitions]),
            W=cat([t.W for t in transitions]),
        )
    
    def __bool__(self):
        return bool(len(self))


from .transition import TransitionBatch, TransitionLoader
from .recorders import BaseRecorder, NStepRecorder

__all__ = (
    'TransitionBatch',
    'TransitionLoader',
    'BaseRecorder'
    'NStepRecorder',
)


import torch
from .torch_base import TorchBaseFunction, TorchProcessor
import tensorflow as tf
from .tf_base import TFBaseFunction, TFProcessor

__all__ = (
    'Q'
)

def Q(*argv, **kwargs):
    if torch.nn.Module in argv[1].__class__.__mro__:
        return Q_torch(*argv, **kwargs)
    if tf.keras.Model in argv[1].__class__.__mro__:
        return Q_tf(*argv, **kwargs)

class Q_torch(TorchBaseFunction):             
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #self.proc = TorchProcessor(self.state_dim, self.action_dim)

    def forward(self, s, a=None, training=True):
        s = self.proc(s)
        a = self.proc(a) if a is not None and type(a) is not torch.Tensor else a
        with torch.inference_mode(mode=not training):
            value = self.model(s, a) if a is not None else self.model(s)
        return value

class Q_tf(TFBaseFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #self.proc = TFProcessor(self.state_dim, self.action_dim)
    
    def call(self, s, a=None, training=True):
        s = self.proc(s)
        a = self.proc(a) if a is not None and type(a.shape) is not tf.TensorShape else a
        value = self.model(s, a, training=training) if a is not None else self.model(s, training=training)
        return value


import wandb

__all__ = (
    'WandbRL'
)

class WandbRL:
    def __init__(self, mode="online", **kwargs):
        if wandb.run:
            wandb.finish() # End previous run
        self.experiment = wandb.init(mode=mode, **kwargs)

    def log_metrics(self, metrics, step):
        self.experiment.log(metrics, step=step)
    
    def log_histogram(self, data, step, bins):
        for key, item in data.items():
            self.experiment.log( {key:wandb.Histogram(item, num_bins=bins)},
                                 step=step )

    def log_model(self, data:dict):
        pass
    
    def log_image(self, data, step):
        raise NotImplementedError
    


import numpy as np
import torch
from accelerate import Accelerator
import copy
from torch.utils.data import DataLoader
from .other import _set_dimensions


class TorchBaseFunction(torch.nn.Module):
                            
    def __init__(self, env, model, optimizer, optim_args={}, trainer_args={}):
        super().__init__()
        self.env = env
        self.model = model
        self.optimizer = optimizer(self.model.parameters(), **optim_args)
        self.optim_args = optim_args
        _set_dimensions(self, env)
        self.accelerator = Accelerator()
        self.proc = TorchProcessor(self.state_dim, self.action_dim, device=self.accelerator.device)
        self.compile()

    def forward(self, s, a=None, training=True):
        raise NotImplementedError

    def soft_update(self, other, tau):
        for old_param, new_param in zip(self.model.parameters(), other.model.parameters()):
            old_param.data.copy_(tau*new_param.data + (1.0-tau)*old_param.data)

    def copy(self, deep=False):
        if deep:
            model_copy = copy.deepcopy(self.model)
            optimizer = self.optimizer
            return self.__class__(self.env, model_copy, type(optimizer), optim_args=self.optim_args)
        return copy.copy(self)
    
    def compile(self):
        self.model, self.optimizer = self.accelerator.prepare(self.model, self.optimizer)
        #setattr(self.proc, "device", self.accelerator.device)
        #self.proc.device = self.accelerator.device

    def _clear(self):
        optimizer = self.optimizer
        optimizer.zero_grad()
        return optimizer, None

    def _gradients(self, loss, **kwargs):
        self.accelerator.backward(loss)
    
    def _apply_grads(self, optimizer, grads=None):
        optimizer.step()

    def _update(self, transition_batch, **kwargs):
        #for tb in self.accelerator.prepare( DataLoader([transition_batch]) ): pass
        loss = self.train_step(transition_batch, **kwargs)
        return loss


class TorchProcessor:
    def __init__(self, state_dim, action_dim, device=None):
        self.state_dim = state_dim
        self.action_dim = action_dim
        if device is None:
            device = torch.device("cpu")
        self.device = device

    def __call__(self, arg, vectorize=False):
        assert type(arg) is np.ndarray, "Processor argument should be numpy array"
        arg = torch.from_numpy(arg).float().to(self.device)
        arg = arg.view(-1, arg.shape[-1]) if self.state_dim in arg.shape or self.action_dim in arg.shape or vectorize else arg
        return arg
    
    def invert(self, arg:torch.Tensor):
        return arg.cpu().numpy()

    def max(self, target, dim, **kwargs):
        values, _ = target.max(dim, **kwargs)
        return values

    def multiply(self, target, *others):
        result = target
        for other in others: result = result.mul(self(other))
        return result

    def gather(self, target, dim, idx):
        idx = torch.from_numpy(idx).view(target.size(0), -1).to(self.device)
        return target.gather(dim, idx)

    def add(self, target, *args):
        result = target
        for arg in args: result += self(arg)
        return result
    
    def reshape(self, target, shape):
        return target.reshape(shape)

    def mean(self, target, dim=None, **kwargs):
        if dim is None:
            return torch.mean(target, **kwargs)
        return torch.mean(target, dim=dim, **kwargs)


import torch
import tensorflow as tf
from .base import Objective
from ..core.other import _bind_method

class DeterministicPG(Objective):
    def __init__(self, pi, q):
        super().__init__()
        self.pi = pi
        self.q = q
        self.proc = pi.proc

    def flush(self):
        self.q._clear()
        return self.pi._clear()

    def compute_loss(self, transition_batch):
        Apred = self.pi(transition_batch.S)
        Q_SApred_neg = -self.q(transition_batch.S, Apred)
        return self.proc.mean(Q_SApred_neg)
    
    def process_grads(self, grads):
        # Do nothing
        return grads

    def compute_grads(self, loss, **kwargs):
        return self.pi._gradients(loss, **kwargs)

    def update_step(self, transition_batch, idx=None):
        optimizer, tape = self.flush()
        loss = self.compute_loss(transition_batch)
        grads = self.compute_grads(loss, tape=tape)
        grads = self.process_grads(grads)
        self.pi._apply_grads(optimizer, grads)
        return loss
    
    def update(self, transition_batch):
        # Perform training step
        loss = self.update_step(transition_batch)
        return {f"{self.__class__.__name__}/loss": loss}


from abc import ABC, abstractmethod

class RLmodel(ABC):
    @abstractmethod
    def gradients(self, loss):
        pass
    
    @abstractmethod
    def update(self, grads):
        pass

    @abstractmethod
    def clear(self):
        # clear gradients
        pass


import numpy as np

__all__ = (
    'TransitionBatch',
    'TransitionLoader'
)


class TransitionBatch:
    def __init__(self, S, A, R, Done, S_next, Gamma, W=None, idx=None):
        self.S = S
        self.A = A
        self.R = R
        self.Done = Done
        self.S_next = S_next
        self.Gamma = Gamma
        self.W = np.ones_like(R) if W is None else W
        self.idx = np.arange(R.shape[0], dtype='int32') if idx is None else idx
    
    @classmethod
    def from_single(cls, s, a, r, done, s_next, gamma, w=1, idx=None):
        return cls(
            S=single_to_batch(s),
            A=single_to_batch(a),
            R=single_to_batch(r),
            Done=single_to_batch(done),
            S_next=single_to_batch(s_next),
            Gamma = single_to_batch(gamma),
            W=single_to_batch(float(w)),
            idx=single_to_batch(idx) if idx is not None else None
        )
    
    @property
    def batch_size(self):
        return np.shape(self.R)[0]
    
    def __len__(self):
        return self.batch_size

    # @property
    # def shape(self):
    #     return (self.batch_size, )

    def to_singles(self):
        if self.batch_size == 1:
            yield self
            return
        
        zipped = zip(self.S, self.A, self.R, self.Done, self.S_next, self.W)

        for s,a,r,done,s_next,w in zipped:
            yield TransitionBatch.from_single(
                s=s, a=a, r=r, done=done, s_next=s_next, w=w
            ) 
    
def single_to_batch(x):
    return np.expand_dims(x, axis=0)


class TransitionLoader:
    def __init__(self, transitions:TransitionBatch, batch_size=None, shuffle=False):
        self.batch_size = batch_size if batch_size is not None else len(transitions)
        self.num_batches = int(np.ceil(len(transitions) / self.batch_size))
        if self.batch_size == len(transitions):
            self.data = [transitions]
        else:
            # Implement mini-batching of transition batch
            pass
        self._index = 0
    
    def __iter__(self):
        return self
    
    def __next__(self):
        if self._index < self.num_batches:
            result = self.data[self._index]
            self._index += 1
            return result
        self._index = 0
        raise StopIteration

    def __len__(self):
        return self.num_batches



import random
import numpy as np

__all__ = (
    'EpsilonGreedy',
    'BoltzmannPolicy'
)

class EpsilonGreedy:
    def __init__(self, epsilon, q, epsilon_min=0.01):
        self.epsilon = epsilon
        self.q = q
        self.epsilon_min = 0.01

    def act(self, s, explore=True):
        if random.random() < self.epsilon and explore:
            #print("Random action ")
            return self.q.env.action_space.sample()
        else:
            #print("Action : ", end=" ")
            qs = self.q(s, training=False)
            qs = self.q.proc.invert(qs)
            qs = qs.ravel() 
            assert len(qs) == self.q.env.action_space.n
            return np.argmax(qs)

    def update(self, decay=0.995):
        if self.epsilon > self.epsilon_min:
            self.epsilon *= decay

class BoltzmannPolicy:
    def __init__(self):
        pass


import numpy as np
from abc import ABC, abstractmethod
import random
from collections import deque
from ..recorder import TransitionBatch


__all__ = (
    'BaseBuffer',
    'ReplayBuffer'
)

class BaseBuffer(ABC):
    @abstractmethod
    def add(self, transition_batch):
        pass
    @abstractmethod
    def sample(self, batch_size=32):
        pass
    @abstractmethod
    def clear(self):
        pass
    @abstractmethod
    def __len__(self):
        pass
    @abstractmethod
    def __bool__(self):
        pass
    @abstractmethod
    def __iter__(self):
        pass

class ReplayBuffer(BaseBuffer):
    def __init__(self, capacity, random_seed=None):
        self.capacity = int(capacity)
        random.seed(random_seed)
        self.random_state = random.getstate()
        self.clear() 

    def clear(self):
        self.storage = deque(maxlen=self.capacity)
        self.index = 0

    def add(self, transition_batch):
        transition_batch.idx = np.arange(self.index, self.index + transition_batch.batch_size)
        self.index += transition_batch.batch_size
        self.storage.extend(transition_batch.to_singles())

    def sample(self, batch_size=32):
        # sandwich sample in between setstate/getstate in case global random state was tampered with
        random.setstate(self.random_state)
        transitions = random.sample(self.storage, batch_size)
        self.random_state = random.getstate()
        cat = lambda x: np.concatenate(x, axis=0)
        return TransitionBatch(
            S=cat([t.S for t in transitions]),
            A=cat([t.A for t in transitions]),
            R=cat([t.R for t in transitions]),
            Done=cat([t.Done for t in transitions]),
            S_next=cat([t.S_next for t in transitions]),
            Gamma=cat([t.Gamma for t in transitions]),
            W=cat([t.W for t in transitions]),
        )

    def __len__(self):
        return len(self.storage)

    def __bool__(self):
        return bool(len(self))

    def __iter__(self):
        return iter(self.storage)


from abc import ABC, abstractmethod

from conrl.core.tf_base import TFBaseFunction
from .torch_base import TorchBaseFunction
import torch

__all__ = (
    'Policy', 'StochasticPolicy'
)

class Policy:
    def __new__(cls, *args, model=None, **kwargs):
        if torch.nn.Module in model.__class__.__mro__:
            #print("Torch policy instantiated")
            return Policy_torch(*args, model=model, **kwargs)
        elif tf.keras.Model in model.__class__.__mro__:
            #print("TF policy instantiated")
            return Policy_tf(*args, model=model, **kwargs)
        else:
            return None

class StochasticPolicy:
    def __new__(cls, *args, model=None, **kwargs):
        if torch.nn.Module in model.__class__.__mro__:
            print("Torch stochastic policy instantiated")
            return StochasticPolicy_torch(*args, model=model, **kwargs)
        elif tf.keras.Model in model.__class__.__mro__:
            print("TF stochastic policy instantiated")
            return StochasticPolicy_tf(*args, model=model,  **kwargs)
        else:
            return None    

class Policy_torch(TorchBaseFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, s, training=True):
        s = self.proc(s)
        with torch.inference_mode(mode=not training):
            a = self.model(s)
        return a

    def act(self, s):
        a = self(s, training=False)
        return self.proc.invert(a).ravel()


class Policy_tf(TFBaseFunction):
    def __init__(self, env, model=None):
        super().__init__(*args, **kwargs) 

    def call(self, s, training=True):
        pass

    def act(self, s):
        a = self(s, training=False)
        return self.proc.invert(a).ravel()



class StochasticPolicy_torch:
    def __init__(self) -> None:
        """TODO"""
        pass



class StochasticPolicy_tf:
    def __init__(self) -> None:
        """TODO"""
        pass


import gym
from .base import Objective
from ..core.other import _bind_method

class TDLearning(Objective):
    def __init__(self, q, q_targ=None, pi_targ = None, loss_function=None):
        self.q = q
        self.q_targ = q_targ
        self.pi_targ = pi_targ
        self.loss_function = loss_function
        self.proc = q.proc

    def flush(self):
        return self.q._clear()
        
    def process_grads(self, grads):
        # Do nothing
        return grads

    def update_step(self, transition_batch, idx=None):
        optimizer, tape = self.flush()
        loss = self.compute_loss(transition_batch)
        grads = self.compute_grads(loss, tape=tape)
        grads = self.process_grads(grads)
        self.q._apply_grads(optimizer, grads)
        return loss

    def update(self, transition_batch):
        # Perform training step
        loss = self.update_step(transition_batch)
        return {f"{self.__class__.__name__}/loss": loss}


class QLearning(TDLearning):
    def __init__(self, q, q_targ=None, pi_targ=None, loss_function=None):
        super().__init__(q, q_targ=q_targ, pi_targ=pi_targ, loss_function=loss_function)
        assert isinstance(self.q.env.action_space, gym.spaces.Discrete)

    def compute_target(self, transition_batch):
        Q_targets_next =  self.q_targ(transition_batch.S_next, training=False) 
        Q_targets_next = self.proc.max(Q_targets_next, dim=1)
        Q_targets = self.proc.multiply(Q_targets_next, 1-transition_batch.Done, transition_batch.Gamma)
        Q_targets = self.proc.add( Q_targets, transition_batch.R )
        return self.proc.reshape(Q_targets, (-1,1))

    def compute_loss(self, transition_batch):
        targets = self.compute_target(transition_batch)
        Q_expected = self.q(transition_batch.S)
        Q_expected = self.q.proc.gather(Q_expected, dim=1, idx=transition_batch.A)
        loss = self.loss_function(Q_expected, targets)    
        return loss

    def compute_grads(self, loss, **kwargs):
        return self.q._gradients(loss, **kwargs)

    

class QLearningwithPolicy(TDLearning):
    def __init__(self, q, q_targ=None, pi_targ=None, loss_function=None):
        super().__init__(q, q_targ=q_targ, pi_targ=pi_targ, loss_function=loss_function)
    
    def compute_target(self, transition_batch):
        A_next = self.pi_targ(transition_batch.S_next, training=False)
        Q_sa_next = self.q_targ(transition_batch.S_next, A_next, training=False).ravel()
        Q_targets = self.proc.multiply(Q_sa_next, 1-transition_batch.Done, transition_batch.Gamma)
        Q_targets = self.proc.add( Q_targets, transition_batch.R )
        return self.proc.reshape(Q_targets, (-1,1))

    def compute_loss(self, transition_batch):
        targets = self.compute_target(transition_batch)
        Q_expected = self.q(transition_batch.S, transition_batch.A)
        assert Q_expected.shape == targets.shape
        loss = self.loss_function(Q_expected, targets)
        return loss

    def compute_grads(self, loss, **kwargs):
        return self.q._gradients(loss, **kwargs)


from gym import spaces
from gym.spaces.discrete import Discrete

def _set_dimensions(obj, env):
    if isinstance(env.observation_space, spaces.Box) or isinstance(env.observation_space, spaces.MultiBinary):
        obj.state_dim = env.observation_space.shape[-1]
    else:
        print("Cannot figure out state size : Think a little")
    
    if isinstance(env.action_space, spaces.Box):
        obj.action_dim = env.action_space.shape[-1]
    elif isinstance(env.action_space, spaces.Discrete):
        obj.action_dim = env.action_space.n
    else:
        print("Cannot figure out action size : Think a little")

def _bind_method(instance, func, as_name=None):
    """
    Bind the function *func* to *instance*, with either provided name *as_name*
    or the existing name of *func*. The provided *func* should accept the 
    instance as the first argument, i.e. "self".
    """
    if as_name is None:
        as_name = func.__name__
    bound_method = func.__get__(instance, instance.__class__)
    setattr(instance, as_name, bound_method)




from abc import ABC, abstractmethod

class Objective(ABC):
    @abstractmethod
    def flush(self): # clears autograd memory
        pass

    @abstractmethod
    def compute_loss(self, transition_batch):
        pass

    @abstractmethod
    def compute_grads(self, loss):
        pass

    @abstractmethod
    def process_grads(self, grads, transition_batch):
        pass

    @abstractmethod
    def update_step(self, transition_batch):
        pass

    @abstractmethod
    def update(self, transition_batch):
        pass


import numpy as np
import time
import gym
from collections import deque
from tqdm.notebook import tqdm

__all__ = (
    'MonitorStats',
)

class StreamingSample:
    def __init__(self, maxlen, random_seed=None):
        self._deque = deque(maxlen=maxlen)
        self._count = 0
        self._rnd = np.random.RandomState(random_seed)

    def reset(self):
        self._deque = deque(maxlen=self.maxlen)
        self._count = 0

    def append(self, obj):
        self._count += 1
        self._deque.append(obj)

    @property
    def values(self):
        return list(self._deque)
    @property
    def maxlen(self):
        return self._deque.maxlen

    def __len__(self):
        return len(self._deque)

    def __bool__(self):
        return bool(self._deque)


class MonitorStats(gym.Wrapper):
    def __init__(
            self, env,
            logger,
            log_all_metrics=False,
            smoothing_window=10,
            **logger_kwargs):

        super().__init__(env)
        self.log_all_metrics = log_all_metrics
        self.smoothing_window = smoothing_window
        self.logger = logger(**logger_kwargs)
        self.reset_global()

    def reset_global(self):
        self.T = 0
        self.ep = 0
        self.t = 0
        self.G = 0.0
        self.avg_G = 0.0
        self._n_avg_G = 0.0
        self._ep_starttime = time.time()
        self._ep_metrics = {}
        self._ep_actions = StreamingSample(maxlen=1000)
        self._scores = deque(maxlen=self.smoothing_window)
        self._period = {'T': {}, 'ep': {}}

    def reset(self):
        # write logs from previous episode:
        if self.ep:
            self._scores.append(self.G)
            self._write_episode_logs()
            tqdm.write(f"\rEpisode: {self.ep} | Timesteps: {self.t} |\
                    Avg_reward: {self.avg_r} | Returns: {self.G}",
                    )
        # increment global counters:
        self.T += 1
        self.ep += 1
        # reset episodic counters:
        self.t = 0
        self.G = 0.0
        self._ep_starttime = time.time()
        self._ep_metrics = {}
        self._ep_actions.reset()

        return self.env.reset()

    @property
    def dt_ms(self):
        if self.t <= 0:
            return np.nan
        return 1000 * (time.time() - self._ep_starttime) / self.t

    @property
    def avg_r(self):
        if self.t <= 0:
            return np.nan
        return self.G / self.t

    @property
    def smooth_G(self):
        return np.mean(self._scores)

    def step(self, a):
        self._ep_actions.append(a)
        s_next, r, done, info = self.env.step(a)
        if info is None:
            info = {}
        info['monitor'] = {'T': self.T, 'ep': self.ep}
        self.t += 1
        self.T += 1
        self.G += r

        return s_next, r, done, info

    def record_metrics(self, metric:dict):
        self.logger.log_metrics(metric, step=self.T)

    def _write_episode_logs(self):
        if self.logger is not None:
            metrics = {
                'episode/episode': self.ep,
                'episode/avg_reward': self.avg_r,
                'episode/return': self.G,
                'episode/smooth_return': self.smooth_G,
                'episode/steps': self.t,
                'episode/avg_step_duration_ms': self.dt_ms}
            self.logger.log_metrics(
                                    metrics, 
                                    step=self.T )
            if self._ep_actions:
                if isinstance(self.action_space, gym.spaces.Discrete):
                    bins = np.arange(self.action_space.n + 1)
                else:
                    bins = 64 # Default in wandb histogram 
                self.logger.log_histogram({"episode/actions":self._ep_actions.values}, 
                                            step=self.T, bins=bins)


import numpy as np
import tensorflow as tf
from tensorflow import keras
import copy
from .other import _set_dimensions


class TFBaseFunction(keras.Model):
    
    def __init__(self, env, model, optimizer, optim_args={}, trainer_args={}):
        super().__init__()
        self.env = env
        self.model = model
        self.optimizer = optimizer(**optim_args)
        self.optim_args = optim_args
        _set_dimensions(self, env)
        self.proc = TFProcessor(self.state_dim, self.action_dim)
        super().compile()

    def call(self, s, a=None, training=True):
        raise NotImplementedError

    def soft_update(self, other, tau):
        updated_params = [tau*new_param + (1.0-tau)*old_param \
                    for old_param, new_param in zip(self.model.get_weights(), other.model.get_weights())]
        self.model.set_weights(updated_params)

    def copy(self, deep=False):
        if deep:
            model_copy = copy.deepcopy(self.model)
            optimizer = self.optimizer
            return self.__class__(self.env, model_copy, type(optimizer), optim_args=optimizer.defaults)
        return copy.copy(self)
    
    def _clear(self):
        optimizer = self.optimizer
        tape = tf.GradientTape()
        tape._push_tape()
        return optimizer, tape

    def _gradients(self, loss, **kwargs):
        tape = kwargs["tape"]
        tape._pop_tape()
        grads = tape.gradient(loss, self.model.trainable_weights)
        return grads

    def _apply_grads(self, optimizer, grads=None):
        optimizer.apply_gradients(zip(grads, self.model.trainable_weights))

    def _update(self, transition_batch, **kwargs):
        loss = self.train_step(transition_batch, **kwargs)
        return loss


class TFProcessor:
    def __init__(self, state_dim, action_dim):
        self.state_dim = state_dim
        self.action_dim = action_dim

    def __call__(self, arg, vectorize=False, **kwargs):
        #assert type(arg) is np.ndarray, "Processor argument should be numpy array"
        arg = tf.convert_to_tensor(arg, dtype=tf.float32) if type(arg) is np.ndarray else arg
        arg = tf.reshape(arg, (-1, arg.shape[-1])) if self.state_dim in arg.shape \
                or self.action_dim in arg.shape or vectorize else arg
        return arg
    
    def invert(self, arg):
        return arg.numpy()

    def max(self, target, dim, **kwargs):
        values = tf.math.reduce_max(target, axis=dim, **kwargs)
        return values

    def multiply(self, target, *others):
        result = target
        for other in others: result = tf.math.multiply(result, other)
        return result

    def gather(self, target, dim, idx):
        idx = idx.reshape(target.shape[0], -1)
        return tf.gather(target, idx, batch_dims=dim)

    def add(self, target, *args):
        result = target
        for arg in args: result += self(arg)
        return result

    def reshape(self, target, shape):
        return tf.reshape(target, shape)

    def mean(self, target, dim, **kwargs):
        return tf.math.reduce_mean(target, axis=dim, **kwargs)


import numpy as np
import torch
import pytorch_lightning as pl
import copy
from .other import _set_dimensions
from ..recorder import TransitionLoader

class TorchBaseFunction(pl.LightningModule):
    
    # done for strong referencing
    trainer = pl.Trainer(
                         max_epochs=1,
                         min_epochs=1, 
                         logger=False, 
                         checkpoint_callback=False,
                         weights_summary=None,
                         progress_bar_refresh_rate=0,
                         num_sanity_val_steps=0,
                         auto_select_gpus=True,
                         #gpus=1,
                         accelerator="auto"
                        )
                            
    def __init__(self, env, model, optimizer, optim_args={}, trainer_args={}):
        super().__init__()
        self.env = env
        self.model = model
        self.optimizer = optimizer
        self.optim_args = optim_args
        self.automatic_optimization = False
        _set_dimensions(self, env)
        self.proc = TorchProcessor(self.state_dim, self.action_dim)

    def forward(self, s, a=None, training=True):
        raise NotImplementedError

    def soft_update(self, other, tau):
        for old_param, new_param in zip(self.model.parameters(), other.model.parameters()):
            old_param.data.copy_(tau*new_param.data + (1.0-tau)*old_param.data)

    def copy(self, deep=False):
        if deep:
            model_copy = copy.deepcopy(self.model)
            optimizer = self.optimizer
            return self.__class__(self.env, model_copy, type(optimizer), optim_args=optimizer.defaults)
        return copy.copy(self)
    
    def configure_optimizers(self):
        return self.optimizer(self.parameters(), **self.optim_args)

    # def _setup(self, transition_batch):
    #     self.transition_batch = transition_batch

    def _clear(self):
        optimizer = self.optimizers()
        optimizer.zero_grad()
        return optimizer, None

    def _gradients(self, loss, **kwargs):
        self.manual_backward(loss)
    
    def _apply_grads(self, optimizer, grads=None):
        optimizer.step()

    def _update(self, transition_batch):
        dataloader = TransitionLoader(transition_batch)
        TorchBaseFunction.trainer.fit(self, dataloader)
        if "loss" not in TorchBaseFunction.trainer.logged_metrics.keys():
            raise KeyError("Use self.log('loss', value) in training step")
        return self.trainer.logged_metrics["loss"]


class TorchProcessor:
    def __init__(self, state_dim, action_dim):
        self.state_dim = state_dim
        self.action_dim = action_dim

    def __call__(self, arg, vectorize=False):
        assert type(arg) is np.ndarray, "Processor argument should be numpy array"
        arg = torch.from_numpy(arg).float()
        arg = arg.view(-1, arg.shape[-1]) if self.state_dim in arg.shape or self.action_dim in arg.shape or vectorize else arg
        return arg
    
    def invert(self, arg:torch.Tensor):
        return arg.numpy()

    def max(self, target, dim, **kwargs):
        values, _ = target.max(dim, **kwargs)
        return values

    def multiply(self, target, *others):
        result = target
        for other in others: result = result.mul(self(other))
        return result

    def gather(self, target, dim, idx):
        idx = torch.from_numpy(idx).view(target.size(0), -1)
        return target.gather(dim, idx)

    def add(self, target, *args):
        result = target
        for arg in args: result += self(arg)
        return result
    
    def reshape(self, target, shape):
        return target.reshape(shape)

